import os as _os
import pickle as _pickle

import tensorflow as _tf
import keras as _keras

import so_ml_tools as _soml


def save_model(model: _tf.keras.Model, filepath: str ="./models/model", save_format: str ="tf") -> None:
    """
    This method will call the default model.save method.

    :param model: the model to save
    :param filepath: String or PathLike, path to the file to save the model to.
    :param save_format: The format to store the data (tf/h5)
    """
    model.save(filepath=filepath, save_format=save_format)


def load_model(filepath="./models/model", custom_objects=None) -> _keras.Model:
    """
    This method will call the default load_model method.

    Important: When loading a model containing TensorFlow Hub layers you will need to provide this
    as a custom object:

    custom_objects={"KerasLayer": hub.KerasLayer}

    However, this is not needed when the model was saved in the SaveModel format since these
    details are present in the saved state.

    :param filepath: String or PathLike, path to the file to save the model to.
    :param custom_objects: custom objects for example KerasLayer from TensorFlow Hub
    :return: the loaded model
    """
    return _tf.keras.models.load_model(filepath=filepath, custom_objects=custom_objects)


def save_weights(model: _tf.keras.Model, filepath: str = "./models/model", save_format: str ="tf") -> None:
    """
    Saves only the weights of the model (see model.save_weights )

    :param model:
    :param filepath:
    :param save_format:
    """
    path = _os.path.dirname(filepath)
    if path:
        _os.makedirs(path, exist_ok=True)
    model.save_weights(filepath, save_format=save_format)


def load_weights(model: _tf.keras.Model, filepath: str = "./models/model") -> None:
    """
    Loads the weights into the given model.

    Disables trainable on the model and on all the layers, then loads the weights and
    restores the trainable configuration on the model and weights.

    :param model: the model to load the weights into
    :param filepath: String or PathLike, path to the file to save the weights
                to. When saving in TensorFlow format, this is the prefix used
                for checkpoint files (multiple files are generated). Note that
                the '.h5' suffix causes weights to be saved in HDF5 format.
    """

    # Collect all layers that are trainable
    trainable_layer_names = _soml.tf.model.inspect.collect_layer_names(model, recursive=True, include_trainable=True,
                                                                     include_non_trainable=False)

    # Disable training completely
    _soml.tf.model.layer.set_trainable_on_layers(model, trainable=False)

    # Load the weights
    model.load_weights(filepath)

    _soml.tf.model.layer.set_trainable_on_layers(model, layer_names=trainable_layer_names, trainable=True)


def save_model_alt(model: _tf.keras.Model, name: str, directory="./models", format='h5') -> None:
    """
    Alternative solution to saving a model since the default implementation has issues with augmentation layers.

    Use in conjunction with load_model_alt, this method will create two files, one .pkl file
    with the configuration of the model, and a (for example .h5) file containing the weights of the model.

    :param model: The model to save
    :param directory: The target directory to save the model to
    :param name: The name of the model, used for filenames, use without extension
    :param format: The format to store the data (tf/h5)
    """
    config = model.get_config()

    # Check if target directory exists, if not, create it
    _os.makedirs(directory, exist_ok=True)

    config_file = _os.path.join(directory, name + '.pkl')
    with open(config_file, 'wb') as fp:
        _pickle.dump(config, fp)

    model_file = _os.path.join(directory, name + '.' + format)
    model.save_weights(model_file, save_format=format)


def load_model_alt(name: str, directory="./models", format='h5'):
    """
    Alternative solution to saving a model since the default implementation has issues with augemntation layers.

    This method will create a new model based on the configuration stored in a .pkl file and the weights stored in
    a (for example .h5) file.

    Use in conjunction with save_model_alt

    :param directory: The target directory to save the model to
    :param name: The name of the model, used for filenames, use without extension
    :param format: The format to store the data (tf/h5)
    :return: the loaded model
    """
    config_file = _os.path.join(directory, name + '.pkl')
    with open(config_file, 'rb') as fp:
        model = _keras.Model.from_config(_pickle.load(fp))
        model_file = _os.path.join(directory, name + '.' + format)
        model.load_weights(model_file)
        return model



import tensorflow as _tf


def load_image_dataset_from_directory(directory: str,
                                      label_mode='categorical',
                                      image_size=(256, 256),
                                      batch_size=32,
                                      class_names=None,
                                      color_mode='rgb',
                                      shuffle=True,
                                      validation_split=None,
                                      seed=None):
    return _tf.keras.utils.image_dataset_from_directory(directory=directory,
                                                        label_mode=label_mode,
                                                        image_size=image_size,
                                                        batch_size=batch_size,
                                                        class_names=class_names,
                                                        color_mode=color_mode,
                                                        shuffle=shuffle,
                                                        validation_split=validation_split,
                                                        seed=seed
                                                        )




import matplotlib.pyplot as _plt
import tensorflow as _tf


def plot_consecutive_histories(histories: list[_tf.keras.callbacks.History], labels: list[str], figsize=(10, 6)):
    """
    Plots (when available), the validation loss and accuracy, training loss and accuracy and learning rate.

    :param histories: the history objects returned from fitting models.
    :param labels: the labels for each history object for seperating the epochs
    :param figsize: figure size (default: (10, 6))
    :return: two graphs, one for loss and one for accuracy
    """
    all_loss_history = []
    all_val_loss_history = []
    all_accuracy_history = []
    all_val_accuracy_history = []
    all_mae_history = []
    all_val_mae_history = []
    all_lr_history = []

    first_epoch = 10000
    last_epoch = -1
    for history in histories:
        first_epoch = min(first_epoch, min(history.epoch))
        last_epoch = max(last_epoch, max(history.epoch))

        all_loss_history = [*all_loss_history, *history.history['loss']]
        if 'val_loss' in history.history:
            all_val_loss_history = [*all_val_loss_history, *history.history['val_loss']]

        if 'accuracy' in history.history:
            all_accuracy_history = [*all_accuracy_history, *history.history['accuracy']]
            all_val_accuracy_history = [*all_val_accuracy_history, *history.history['val_accuracy']]

        if 'mae' in history.history:
            all_mae_history = [*all_mae_history, *history.history['mae']]
            all_val_mae_history = [*all_val_mae_history, *history.history['val_mae']]

        if 'lr' in history.history:
            all_lr_history = [*all_lr_history, *history.history['lr']]

    epoch_labels = range(first_epoch + 1, last_epoch + 2)
    ticks = range(len(epoch_labels))
    _plt.figure(figsize=figsize, facecolor='#FFFFFF')
    _plot_history_graph_line(all_loss_history, label='Training loss', color='#0000FF')
    _plot_history_graph_line(all_lr_history, label='Learning rate', color='#000000', linestyle='dashed')

    if all_val_loss_history:
        _plot_history_graph_line(all_val_loss_history, label='Validation loss', color='#00FF00')

    _plot_history_ends(histories, labels)
    _plt.title('Loss', size=20)
    _plt.xticks(ticks, epoch_labels)
    _plt.xlabel('Epoch', size=14)
    _plt.legend()

    if all_accuracy_history:
        # Start a new figure
        _plt.figure(figsize=figsize, facecolor='#FFFFFF')
        _plot_history_graph_line(all_accuracy_history, label='Training accuracy', color='#0000FF')
        _plot_history_graph_line(all_val_accuracy_history, label='Validation accuracy', color='#00FF00')
        _plot_history_graph_line(all_lr_history, label='Learning rate', color='#000000', linestyle='dashed')
        _plot_history_ends(histories, labels)
        _plt.title('Accuracy', size=20)
        _plt.xticks(ticks, epoch_labels)
        _plt.xlabel('Epoch', size=14)
        _plt.legend()

    if all_mae_history:
        # Start a new figure
        _plt.figure(figsize=figsize, facecolor='#FFFFFF')
        _plot_history_graph_line(all_mae_history, label='Training mae', color='#0000FF')
        _plot_history_graph_line(all_val_mae_history, label='Validation mae', color='#00FF00')
        _plot_history_graph_line(all_lr_history, label='Learning rate', color='#000000', linestyle='dashed')
        _plot_history_ends(histories, labels)
        _plt.title('Mean Absolute Accuracy', size=20)
        _plt.xticks(ticks, epoch_labels)
        _plt.xlabel('Epoch', size=14)
        _plt.legend()


def plot_history(history: _tf.keras.callbacks.History, figsize=(10, 6)):
    plot_consecutive_histories([history], ["Start history"], figsize=figsize)


def _plot_history_ends(histories: list[_tf.keras.callbacks.History], labels: list[str]) -> None:
    """
    Internal method which will plot a vertical line showing where a histories last epoch is visible.

    :param histories: the history objects returned from fitting models.
    :param labels: the labels for each history object for seperating the epochs
    """
    for idx, history in enumerate(histories):
        _plt.plot([min(history.epoch), min(history.epoch)], _plt.ylim(), label=f'{labels[idx]}')


def _plot_history_graph_line(data, label, color, linestyle='solid') -> None:
    """
    Internal method which will plot the information from the fit histroy.

    :param data: the data to plot
    :param label: the label associated with the data
    :param color: color of the line
    :param linestyle: line-style of the line (default: solid)
    """
    if data:
        _plt.plot(data, label=label, color=color, linestyle=linestyle, linewidth=1.5)


import pandas as _pd
import numpy as _np

from sklearn.preprocessing import OneHotEncoder

########################################################################################################################
# General
########################################################################################################################


def one_hot_encode_column(*columns: list[any]) -> (OneHotEncoder, list[any]):
    """
    One-Hot encodes all given columns using all values of all the provided columns.

    Example usage:

    encoder, train_labels_one_hot, val_labels_one_hot, test_labels_one_hot = \
    one_hot_encode_column(
        train_df['target'],
        val_df['target'],
        test_df['target']
    )

    If you get the ValueError:

    all the input arrays must have same number of dimensions, but the array at index 0 has 1 dimension(s) and the array at index 1 has 2 dimension(s)

    Then that means one of the lists passed in has ha different shape then the others.

    :param columns: one or more columns
    :return: first value is the encoder, all other values are in same order as the columns but encoded
    """
    encoder = OneHotEncoder(sparse_output=False)

    # Preprocess the columns
    preprocessed_columns = []
    for column in columns:
        if isinstance(column, _pd.DataFrame) or isinstance(column, _pd.Series):
            column = column.to_numpy()

        # Make sure the column is two-dimensional
        if column.ndim == 1:
            column = column.reshape(-1, 1)

        preprocessed_columns.append(column)

    # Concatenate all columns and fit the encoder
    encoder.fit(_np.concatenate(preprocessed_columns))

    return_values = []
    # Transform each list now individually
    for preprocessed_column in preprocessed_columns:
        return_values.append(encoder.transform(preprocessed_column))

    return_values.insert(0, encoder)
    return return_values


def convert_column_to_type(dataframe: _pd.DataFrame, columns: list[str], dtype=_np.float64,
                           inplace=True) -> _pd.DataFrame:
    """
    Converts the dtype of the given column to the specified dtype.

    :param dataframe: the pd.DataFrame
    :param columns: A `list` of column names
    :param dtype: the newly to assign dtype.
    :param inplace: return a new instance of the DataFrame or adjust the given DataFrame
    :return: see inplace
    """
    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    for column in columns:
        work_df[column] = work_df[column].astype(dtype)

    return work_df


def delete_null_rows(dataframe: _pd.DataFrame, column_names: list[str], inplace=True) -> _pd.DataFrame:
    """
    Deletes all rows containing a null values inside the given column.

    :param dataframe: the pd.DatFrame
    :param column_names: the names of the columns
    :param inplace: return a new instance of the DataFrame (False) or adjust the given DataFrame
    :return: see inplace
    """
    if not type(column_names) == list and column_names is not None:
        column_names = [column_names]

    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    for c in column_names:
        if c in dataframe:
            work_df.drop(work_df[dataframe[c].isnull()].index, inplace=True)
        else:
            print(f"delete_null_rows: Column '{c}' does not exist in dataframe.")

    return work_df


def fill_nan_with_value(dataframe: _pd.DataFrame, column_values: dict, inplace=True) -> _pd.DataFrame:
    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    for c, v in column_values.items():
        if c in dataframe:
            work_df[c].fillna(value=v, inplace=True)
        else:
            print(f"fill_nan_with_value: Column '{c}' does not exist in dataframe.")

    return work_df


def generate_code_ordinal_encoder(dataframe: _pd.DataFrame, column_names: list[str]) -> None:
    for c in column_names:
        if c in dataframe:
            values = dataframe[c].value_counts()
            categories = "', '".join(values)
            print(f"sklearn.preprocessing.OrdinalEncoder(categories=[['{categories}']])")
        else:
            print(f"delete_null_rows: Column '{c}' does not exist in dataframe.")


def delete_rows_not_numeric(dataframe: _pd.DataFrame, column_name: str, inplace=True) -> _pd.DataFrame:
    """
    Deletes all rows containing non-number values inside the given column.

    :param dataframe: the pd.DatFrame
    :param column_name: the column
    :param inplace: return a new instance of the DataFrame (False) or adjust the given DataFrame
    :return: see inplace
    """
    return dataframe.drop(dataframe[_pd.to_numeric(dataframe[column_name], errors='coerce').isna()].index,
                          inplace=inplace)


def drop_columns(dataframe: _pd.DataFrame, column_names: list[str]) -> None:
    """
    Removes (inplace) the given column names from the dataframe.

    :param dataframe: the pd.DatFrame
    :param column_names: the names of the columns to remove
    """
    if not type(column_names) == list and column_names is not None:
        column_names = [column_names]

    for c in column_names:
        if c in dataframe:
            dataframe.drop(c, axis=1, inplace=True)
        else:
            print(f"drop_columns: Column '{c}' does not exist in dataframe.")


def describe(dataframe: _pd.DataFrame, column_names: list[str] = None, round=2):
    # If the column_names argument is not a list then create a list
    if not type(column_names) == list and column_names is not None:
        column_names = [column_names]

    # If we don't have a list of column names then create a histogram for every column.
    if column_names is not None:
        columns = list(dataframe[column_names].columns)
    else:
        columns = list(dataframe.columns)

    data = []
    for c in columns:
        v = dataframe[c]

        _mean = _std = _min = _q25 = _q50 = _q75 = _max = z_min = z_max = _np.NAN
        if _pd.api.types.is_numeric_dtype(v):
            _mean = v.mean()
            _std = v.std()
            _min = v.min()
            _q25 = v.quantile(.25)
            _q50 = v.quantile(.50)
            _q75 = v.quantile(.75)
            _max = v.max()

            # Calculate lower value when Z-Score = -3
            z_min = -3 * _std + _mean

            # Calculate upper value when Z-Score = 3
            z_max = 3 * _std + _mean

        data.append([
            v.name, v.dtype, v.count(), v.isna().sum(), v.nunique(), _mean, _std, z_min, z_max, _min, _q25,
            _q50, _q75, _max
        ])

    print(f"Total number of rows: {len(dataframe)}")
    return _pd.DataFrame(
        columns=["Column", "DType", "NotNull", "Null", "Unique", "Mean", "Std", "Z-Min", "Z-Max", "Min", "25%", "50%",
                 "75%", "Max"], data=data).round(round)



from keras import mixed_precision as _mixed_precision


def set_single_precision_policy() -> None:
    """
    Disable mixed precision by using the float32 policy.
    """
    policy = _mixed_precision.Policy("float32")
    _mixed_precision.set_global_policy(policy)


def set_mixed_precision_policy_for_gpu() -> None:
    """
    Enable mixed precision by using the mixed_float16 policy, use this policy
    for GPU acceleration.

    For details: https://www.tensorflow.org/guide/mixed_precision
    """
    policy = _mixed_precision.Policy("mixed_float16")
    _mixed_precision.set_global_policy(policy)


def set_mixed_precision_policy_for_tpu() -> None:
    """
    Enable mixed precision by using the mixed_bfloat16 policy, use this policy
    for TPU acceleration.

    For details: https://www.tensorflow.org/guide/mixed_precision
    """
    policy = _mixed_precision.Policy("mixed_bfloat16")
    _mixed_precision.set_global_policy(policy)


def get_mixed_precision_policy():
    """
    Returns the current configured mixed-precision policy.
    :return: the current configured mixed-precision policy.
    """
    return _mixed_precision.global_policy()
